﻿using UnityEngine;
using System.Collections;
using HarmonyLib;
using System.Collections.Generic;
using UnityEngine.UI;
using System.IO;
using System.Threading.Tasks;

public class GlassWalls : Mod
{
    private Material glassMaterial;
    private bool isInGame;
    private Harmony harmony;
    private BlockCreator blockCreator;
    private int baseUniqueIndex = 6970;

    public static Item_Base glassFoundationItem;
    public static Item_Base glassFoundationTriangularItem;
    public static Item_Base glassFoundationTriangularMirroredItem;
    public static Item_Base glassFloorItem;
    public static Item_Base glassFloorTriangularItem;
    public static Item_Base glassFloorTriangularMirroredItem;
    public static Item_Base glassWallItem;
    public static Item_Base glassWallHalfItem;
    public static Item_Base glassWallVItem;
    public static Item_Base glassWallSlopeItem;
    public static Item_Base glassWallSlopeInvItem;
    public static Item_Base glassWallWindowItem;
    public static Item_Base glassWallHalfWindowItem;
    public static Item_Base woodWallWindowItem;


    public void Awake()
    {
        isInGame = false;
        StartCoroutine("Init");

        harmony = new Harmony("com.boebie.GlassWalls");
        harmony.PatchAll();
    }

    public override void WorldEvent_WorldLoaded()
    {
        isInGame = true;
        blockCreator = FindObjectOfType<BlockCreator>();

        // TODO -> OccupyingComponent has a function that causes Foundation materials to reset.
        foreach (var glassFoundation in FindObjectsOfType<Block_Glass_Foundation>())
        {
            foreach (var child in glassFoundation.GetComponentsInChildren<MeshRenderer>())
            {
                child.material = glassMaterial;
            }
        }
        
        CreateBuildMenu();
    }

    public override void WorldEvent_WorldUnloaded()
    {
        isInGame = false;
    }

    public IEnumerator Init()
    {
        AssetBundleCreateRequest request = AssetBundle.LoadFromMemoryAsync(GetEmbeddedFileBytes("glasswalls.assets"));
        yield return request;
        AssetBundle assetBundle = request.assetBundle;

        CreateGlassMaterial(assetBundle);
        CreateGlassObjects();

        LogInfo("Mod loaded. Enjoy the view!");

        while (RAPI.GetLocalPlayer() == null)
            yield return new WaitForEndOfFrame();

        isInGame = true;
    }

    void CreateBuildMenu()
    {
        CanvasHelper helper = ComponentManager<CanvasHelper>.Value;
        GameMenu menu = helper.GetMenu(MenuType.BuildMenu);
        Transform menuObject = menu.menuObjects[0].transform;

        CloneBuildMenuItem("Foundation_Tier3", glassFoundationItem, menuObject);
        CloneBuildMenuItem("Foundation_Triangular_Tier3", glassFoundationTriangularItem, menuObject);
        CloneBuildMenuItem("Floor_Tier3", glassFloorItem, menuObject);
        CloneBuildMenuItem("Floor_Triangular_Tier3", glassFloorTriangularItem, menuObject);
        CloneBuildMenuItem("WallTier3", glassWallItem, menuObject);
        CloneBuildMenuItem("HalfWallTier3", glassWallHalfItem, menuObject);
        CloneBuildMenuItem("WallSlopeParent/VerticalParent/WallSlopeTier3", glassWallSlopeItem, menuObject);
        CloneBuildMenuItem("WallSlopeParent_Mirrored/VerticalParent/WallSlopeTier3", glassWallSlopeInvItem, menuObject);
        CloneBuildMenuItem("WallVSlopeTier3", glassWallVItem, menuObject);
        CloneBuildMenuItem("WindowTier3", glassWallWindowItem, menuObject);
        CloneBuildMenuItem("HalfWindowTier3", glassWallHalfWindowItem, menuObject);
    }

    void CloneBuildMenuItem(string name, Item_Base item, Transform buildMenu)
    {
        if (buildMenu.FindChildRecursively(name) == null)
        {
            LogError($"{name} in BuildMenu couldn't be found.");
            return;
        }

        Transform menuItem = buildMenu.FindChildRecursively(name);
        Transform newSelectBlock = Instantiate(menuItem, menuItem.parent);
        Traverse.Create(newSelectBlock.GetComponent<BuildMenuItem_SelectBlock>()).Field("buildableItem").SetValue(item);

        // TODO -> Resize parent UI element
    }

    void CreateGlassMaterial(AssetBundle assetBundle)
    {
        glassMaterial = assetBundle.LoadAsset<Material>("Glass_Mat");
    }

    Item_Base CreateObject(string baseItemName, string uniqueName, ItemInstance_Recipe recipe, int glassQuantity)
    {
        if (ItemManager.GetItemByName(baseItemName) == null)
        {
            LogError($"{baseItemName} item doens't exist.");
            return null;
        }

        Item_Base baseItem = ItemManager.GetItemByName(baseItemName);
        Item_Base glassClone = baseItem.Clone(baseUniqueIndex, uniqueName);
        glassClone.name = glassClone.UniqueName;
        glassClone.settings_Inventory.LocalizationTerm = "";
        glassClone.settings_Inventory.DisplayName = recipe.SubCategory;
        glassClone.settings_Inventory.Description = "Glass, glass, glass.";
        glassClone.settings_recipe = recipe;

        // Set recipe - not working yet, doesn't appear in build menu
        glassClone.settings_recipe.SetRecipe(new CostMultiple[]
        {
            new CostMultiple(new Item_Base[] {ItemManager.GetItemByName("Glass")}, glassQuantity)
        }, 1);

        var prefabsBase = Traverse.Create(glassClone.settings_buildable).Field<Block[]>("blockPrefabs");
        var newPrefabs = new Block[prefabsBase.Value.Length];

        // Thank you again Aidanamite
        var prefabParent = new GameObject("prefabParent").transform;
        prefabParent.gameObject.SetActive(false);
        DontDestroyOnLoad(prefabParent.gameObject);

        for (int i = 0; i < prefabsBase.Value.Length; i++)
        {
            newPrefabs[i] = Instantiate(prefabsBase.Value[i], new Vector3(0, -42069 /*haha very funny*/, 0), new Quaternion(0, 0, 0, 1), prefabParent);

            if (newPrefabs[i].GetComponent<Block_Foundation>() == null)
                newPrefabs[i].gameObject.AddComponent<Block_Glass>();
            else
                newPrefabs[i].gameObject.AddComponent<Block_Glass_Foundation>();

            newPrefabs[i].buildableItem = glassClone;
            newPrefabs[i].name = glassClone.UniqueName;

            foreach (var child in newPrefabs[i].GetComponentsInChildren<MeshRenderer>())
            {
                child.material = glassMaterial;
            }

            // Slightly changing collider to resolve build errors
            if (uniqueName == "Block_Wall_Glass" || uniqueName == "Block_Wall_Glass_Half" || uniqueName == "Block_Wall_Window_Glass" || uniqueName == "Block_Wall_Window_Glass_Half")
            {
                foreach (var child in newPrefabs[i].GetComponentsInChildren<BoxCollider>())
                {
                    child.size -= new Vector3(0.1f, 0f, 0f);
                }
            }
        }

        prefabsBase.Value = newPrefabs;

        foreach (var type in Resources.FindObjectsOfTypeAll<SO_BlockQuadType>())
            if (type.AcceptsBlock(baseItem))
                Traverse.Create(type).Field("acceptableBlockTypes").GetValue<List<Item_Base>>().Add(glassClone);


        RAPI.RegisterItem(glassClone);
        baseUniqueIndex++;
        return glassClone;
    }

    void CreateGlassObjects()
    {
        glassFoundationItem = CreateObject("Block_Foundation_Tier3", "Block_Foundation_Glass", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Glass Foundation", 0), 2);
        glassFoundationTriangularItem = CreateObject("Block_Foundation_Triangular_Tier3", "Block_Foundation_Triangular_Glass", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Triangular Glass Foundation", 0), 1);
        glassFoundationTriangularMirroredItem = CreateObject("Block_Foundation_Triangular_Tier3_Mirrored", "Block_Foundation_Triangular_Glass_Mirrored", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Triangular Glass Foundation", 0), 1);
        glassFloorItem = CreateObject("Block_Floor_Tier3", "Block_Floor_Glass", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Glass Floor", 0), 2);
        glassFloorTriangularItem = CreateObject("Block_Floor_Triangular_Tier3", "Block_Floor_Triangular_Glass", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Triangular Glass Floor", 0), 1);
        glassFloorTriangularMirroredItem = CreateObject("Block_Floor_Triangular_Tier3_Mirrored", "Block_Floor_Triangular_Glass_Mirrored", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Triangular Glass Floor", 0), 1);
        glassWallItem = CreateObject("Block_Wall_Tier3", "Block_Wall_Glass", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Glass Wall", 0), 2);
        glassWallHalfItem = CreateObject("Block_Wall_Tier3_Half", "Block_Wall_Glass_Half", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Half Glass Wall", 0), 1);
        glassWallSlopeItem = CreateObject("Block_Wall_Slope_Tier3", "Block_Wall_Slope_Glass", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Triangular Glass Wall", 0), 1);
        glassWallVItem = CreateObject("Block_Wall_VSlope_Tier3", "Block_Wall_VSlope_Glass", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Pyramid Glass Wall", 0), 1);
        glassWallSlopeInvItem = CreateObject("Block_Wall_Slope_Tier3_Inverted", "Block_Wall_Slope_Glass_Inverted", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Triangular Glass Wall", 0), 1);
        glassWallWindowItem = CreateObject("Block_Wall_Window_Tier3", "Block_Wall_Window_Glass", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Glass Wall Window", 0), 2);
        glassWallHalfWindowItem = CreateObject("Block_Wall_Window_Tier3_Half", "Block_Wall_Window_Glass_Half", new ItemInstance_Recipe(CraftingCategory.Hidden, true, true, "Half Glass Wall", 0), 1);

        blockCreator = FindObjectOfType<BlockCreator>();
    }

    void Update()
    {
        if (isInGame)
        {
            if (blockCreator == null)
            {
                // Probably not necessary
                blockCreator = FindObjectOfType<BlockCreator>();
            }
        }
    }

    void LogError(string message)
    {
        Debug.LogError($"GlassWalls - {message}");
    }

    void LogInfo(string message)
    {
        Debug.Log($"GlassWalls - {message}");
    }

    public class Block_Glass_Foundation : Block_Foundation
    {
        public static List<Block_Glass_Foundation> objects = new List<Block_Glass_Foundation>();
        public void OnBlockPlaced()
        {
            objects.Add(this);
        }

        public void OnDestroy()
        {
            if (objects.Contains(this))
                objects.Remove(this);
        }
    }

    public class Block_Glass : Block
    {
        public static List<Block_Glass> objects = new List<Block_Glass>();
        public void OnBlockPlaced()
        {
            objects.Add(this);
        }

        public void OnDestroy()
        {
            if (objects.Contains(this))
                objects.Remove(this);
        }
    }
}

// Hopefully Aidanamite doesn't mind me using this code :)
static class ExtentionMethods
{
    public static Item_Base Clone(this Item_Base source, int uniqueIndex, string uniqueName)
    {
        Item_Base item = ScriptableObject.CreateInstance<Item_Base>();
        item.Initialize(uniqueIndex, uniqueName, source.MaxUses);
        item.settings_buildable = source.settings_buildable.Clone();
        item.settings_consumeable = source.settings_consumeable.Clone();
        item.settings_cookable = source.settings_cookable.Clone();
        item.settings_equipment = source.settings_equipment.Clone();
        item.settings_Inventory = source.settings_Inventory.Clone();
        item.settings_recipe = source.settings_recipe.Clone();
        item.settings_usable = source.settings_usable.Clone();
        return item;
    }

    public static void SetRecipe(this ItemInstance_Recipe item, CostMultiple[] cost, int amountToCraft = 1)
    {
        Traverse.Create(item).Field("amountToCraft").SetValue(amountToCraft);
        item.NewCost = cost;
    }
}